import "webcomponent-qr-code";

/* These styles are only for the `qr-code.shadowroot` */
const css = `
	svg {
		width: 100%;
		max-height: 100%;
		height: auto;
	}
`;

export default class SpaceQRCode extends HTMLElement {
	static get observedAttributes() {
		return ["value"];
	}
	get value() {
		return this.getAttribute("value");
	}
	attributeChangedCallback(name, oldValue, newValue) {
		if (name === "value" && newValue && this.$qrcode) {
			this.render(newValue);
		}
	}
	connectedCallback() {
		/* the qr code */
		this.$qrcode = document.createElement("qr-code");
		this.$qrcode.addEventListener("click", this.handleQRClick.bind(this));
		this.$qrcode.setAttribute("format", "svg");
		this.$qrcode.setAttribute("modulesize", "5");

		/* can be tabbed with keyboard */
		this.$qrcode.setAttribute("tabindex", "0");

		/* the copy input */
		this.$buttonCopy = document.createElement("button");
		this.$buttonCopy.innerHTML = "Copy URL";
		this.$buttonCopy.title =
			"Copy to the clipboard, the URL address of the current page (the URL also contains all the content).";
		this.$buttonCopy.addEventListener("click", this.copyURL.bind(this));

		/* insert in dom */
		this.append(this.$qrcode);
		this.append(this.$buttonCopy);

		/* render value as attribute */
		if (this.value) {
			this.render(this.value);
		}
	}

	render(value) {
		/* update the value of the qr code */
		this.$qrcode.setAttribute("data", value);
		/* add custom styles to the shadow root,
			 after we update data value, or it overwrittes them */
		const $styles = document.createElement("style");
		$styles.innerHTML = css;
		this.$qrcode.shadowRoot.appendChild($styles);

		this.$buttonCopy.setAttribute("value", value);
	}
	copyURL({ target }) {
		/* copy current URL to clipboard */
		navigator.clipboard.writeText(this.value);

		/* set is-active atr on element for styles */
		target.setAttribute("is-active", true);
		window.setTimeout(() => {
			target.removeAttribute("is-active");
		}, 600);
	}
	handleQRClick() {
		/* const svg = this.$qrcode.shadowRoot.querySelector('svg') */
		/* window.open(`data:image/svg+xml;utf8,${svg}`) */
		if (this.getAttribute("is-active") === "true") {
			this.removeAttribute("is-active");
		} else {
			this.setAttribute("is-active", true);
		}
	}
}
